# Name of point data section in point cloud structure
point = 'vertex'

# Name of the normalized height point attribute
normalized_height = 'normalized_height'

# Name of the intensity point attribute
intensity = 'intensity'

#
point_cloud = 'pointcloud'

#
provenance = 'log'
