"""Tests for the feature_extractor module."""
from .feature_test1 import Test1FeatureExtractor
from .feature_test23 import Test2FeatureExtractor, Test3FeatureExtractor, TestVectorizedFeatureExtractor
from .feature_test_broken import TestBrokenFeatureExtractor
