import unittest


class Test(unittest.TestCase):
    def test_config_test(self):
        """ This is the first test of the project and should pass if test runner is configured correctly. """
        pass
